open Formula

(* TODO incorrect wrt variable capture *)

let rec without_quantifiers : source_quantified_formula ->
  source_unquantified_formula = function
  | `Exists(_, f) | `Forall(_, f) -> without_quantifiers f
  | f -> quantified_to_unquantified f;;

let rec extract_quantifiers_blocks_exists_phase :
    source_quantified_formula -> string list list = function
  | `Exists(quantifiers, f) ->
      quantifiers::(extract_quantifiers_blocks_forall_phase f)
  | `Forall _ -> failwith "bad phase"
  | _ -> []

and extract_quantifiers_blocks_forall_phase :
    source_quantified_formula -> string list list = function
  | `Forall(quantifiers, f) ->
      quantifiers::(extract_quantifiers_blocks_exists_phase f)
  | `Exists _ -> failwith "bad phase"
  | _ -> [];;

let simple_generalize_model (context : Formula_yices.linear_translate_context)
    (model : Formula_yices.smt_model) =
  let atoms = ref [] in
    Hashtbl.iter
       (fun (linear_atomic : linear_atomic) yices_expr ->
          match linear_atomic with
            | `Propositional _ -> failwith "propositional"
            | `Greater_eq expr ->
                let polarity =
                  if ((Q.sgn (eval_linear_expression
                                model.Formula_yices.real_model expr)) >= 0)
                  then `GREATER_EQ
                  else `LESS in
                  atoms := (expr, polarity) :: !atoms
       )
       context.Formula_yices.linear_translate_atomic_map; !atoms;;

let rec generalize_wrt_test (test : 'a list -> bool) (already : 'a list) :
    'a list -> 'a list = function
      | [] -> already
      | h::t ->
	  if test (already @ t)
	  then generalize_wrt_test test already t
	  else generalize_wrt_test test (h::already) t;;

let rec qtest
    (max_generalize : bool)
    (lc : linearization_context)
    (context0 : Formula_yices.dual_linear_translate_context)
    (output_dimension : int)
    (block_widths_and_contexts :
       (int * Formula_yices.linear_translate_context) list)
    (c : (Formula.linear_expression * convex_comparison_op) list) :
    (Formula.linear_expression * convex_comparison_op) list Formula_yices.smt_answer =
  let cf =
    ((Formula_yices.conjuncts_to_conjunction
		 (c :> (Formula.linear_expression * Formula.convex_comparison_op) list))
	      : Formula.linear_atomic Formula.conjunction
	      :> linear_unquantified_formula) in
  match block_widths_and_contexts with
    | [] (* leaf unquantified formula *) ->
	let translated = Formula_yices.translate_linear_formula
	     context0.Formula_yices.positive_context cf in
	Yices.push context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	Yices.assert_expr context0.Formula_yices.positive_context.Formula_yices.linear_translate_context translated;
	let model =
	  if max_generalize
	  then Formula_yices.get_a_generalized_model lc context0
	    (* TODO restructurer ce bordel en deux fonctions, une qui retourne true/false et l'autre qui sort un modele *)
	  else (match Formula_yices.linear_solve context0.Formula_yices.positive_context with
		  | Formula_yices.UNSATISFIABLE -> Formula_yices.UNSATISFIABLE
		  | Formula_yices.SATISFIABLE _ -> Formula_yices.SATISFIABLE []) in
	  Yices.pop context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	  (model :> (Formula.linear_expression * convex_comparison_op) list Formula_yices.smt_answer)

    | (block_width, context)::rest ->
	let translated = Formula_yices.translate_linear_formula context cf in
	Yices.push context.Formula_yices.linear_translate_context;
	Yices.assert_expr context.Formula_yices.linear_translate_context
          translated;
	let perhaps_numeric_model = Formula_yices.linear_solve context in
	  Yices.pop context.Formula_yices.linear_translate_context;

	  (match perhaps_numeric_model with
	     | Formula_yices.UNSATISFIABLE ->
		 Formula_yices.UNSATISFIABLE
	     | Formula_yices.SATISFIABLE c2 ->
		 let model = simple_generalize_model context c2 in
		 (match qtest true lc context0 (output_dimension + block_width)
		    rest model with
		      | Formula_yices.UNSATISFIABLE ->
			  Formula_yices.SATISFIABLE
			    (if max_generalize
			     then (generalize_wrt_test
			       (fun tested_model ->
				  match qtest false lc context0 (output_dimension + block_width) rest tested_model with
				    | Formula_yices.UNSATISFIABLE -> true
				    | Formula_yices.SATISFIABLE _ -> false)
			       [] model)
			     else model)
		      | Formula_yices.SATISFIABLE model2 ->
 			  let to_be_removed =
			    Formula_yices.conjuncts_to_formula
			      (Formula_yices.conjuncts_inject
				 (Projection.project_minimize_constraints
 				    model2 output_dimension)) in
			  (* print_string "to be removed: ";
			  Formula_math.print_source_unquantified_formula
			    stdout
			    (unlinearize_unquantified_formula lc
			       to_be_removed);
			    print_newline(); *)
			  Yices.assert_expr
			    context.Formula_yices.linear_translate_context
			    (Formula_yices.translate_linear_formula context
			    (`Not to_be_removed));
			  qtest max_generalize lc context0 output_dimension
			    block_widths_and_contexts c));;

let negate_appropriately n f =
  if (n mod 2)==1
  then mk_not f
  else f;;

let linearization_context_from_array (a : string array) =
  let hashtable = Hashtbl.create (Array.length a) in
    Array.iteri (fun i x -> Hashtbl.add hashtable x i) a;
    {
      trans_var_table = hashtable;
      trans_nr_vars = Array.length a;
      trans_var_rev_table = a
    };;

let add_blocking_to_root
    (context0 : Formula_yices.dual_linear_translate_context)
    (block_widths_and_contexts :
       (int * Formula_yices.linear_translate_context) list)
    (blocking_clause : linear_unquantified_formula) =
  match block_widths_and_contexts with
    | [] -> Yices.assert_expr
	context0.Formula_yices.positive_context.Formula_yices.linear_translate_context
	  (Formula_yices.translate_linear_formula
	     context0.Formula_yices.positive_context blocking_clause)
    | (_, context)::_ ->
	Yices.assert_expr context.Formula_yices.linear_translate_context
	  (Formula_yices.translate_linear_formula context blocking_clause);;
    
let rec elim_to_dnf (lc : linearization_context)
    (context0 : Formula_yices.dual_linear_translate_context)
    (output_dimension : int)
    (block_widths_and_contexts :
       (int * Formula_yices.linear_translate_context) list)
    (already : (Formula.linear_expression * convex_comparison_op) list list) =
  match qtest true lc context0 output_dimension block_widths_and_contexts [] with
    | Formula_yices.UNSATISFIABLE -> already
    | Formula_yices.SATISFIABLE model ->
 	let to_be_removed =
	  Formula_yices.conjuncts_to_formula model in
	  add_blocking_to_root context0 block_widths_and_contexts
	    (`Not  to_be_removed);
	  elim_to_dnf lc context0 output_dimension block_widths_and_contexts
	    (model::already);;

let eliminate modulo_theory (f : source_quantified_formula) :
    source_unquantified_formula =
  let unq = without_quantifiers f
  and free_variables = StringSet.elements
    (collect_free_variables_in_formula
       StringSet.empty StringSet.empty f) in
  let output_dimension = List.length free_variables in
  let quantifier_blocks = extract_quantifiers_blocks_forall_phase f in
    
  let lc = linearization_context_from_array
    (Array.of_list (List.concat (free_variables::quantifier_blocks))) in

  let context0 = Formula_yices.mk_dual_context lc
    (linearize_unquantified_formula lc modulo_theory)
    (linearize_unquantified_formula lc 
       (negate_appropriately (List.length quantifier_blocks) unq)) in

  let block_widths_and_contexts =
    List.map
      (fun block -> (List.length block),
	 Formula_yices.mk_linear_translate_context
           (Yices.mk_context ()) lc) quantifier_blocks in
  let dnf = elim_to_dnf lc context0 output_dimension block_widths_and_contexts [] in
    unlinearize_unquantified_formula lc
      (`Or
	 ((List.map Formula_yices.conjuncts_to_conjunction dnf)
	 : Formula.linear_atomic Formula.conjunction list
	 :> linear_unquantified_formula list));;
